import pytest
import os
import tempfile
import numpy as np
np.random.seed(123)

import dyneusr as ds
from dyneusr.core import DyNeuGraph
from dyneusr.mapper.wrappers import KMapperWrapper


class TestDyNeuGraph:
    """ Simple tests for DyNeuGraph
    """
    def test_init(self):
        mapper = KMapperWrapper(verbose=0, memory=None)
        data = np.random.rand(100, 3)
        graph = mapper.fit_map(data)

        dG = DyNeuGraph(G=graph)

        assert dG.A.shape[0] == len(graph['nodes'])
        assert dG.TCM.shape[0] == data.shape[0]
        return 


    def test_init_with_y(self):
        mapper = KMapperWrapper(verbose=0, memory=None)
        data = np.random.rand(100, 3)
        graph = mapper.fit_map(data)
        y = mapper.lens_

        dG = DyNeuGraph(G=graph, y=y)

        assert dG.A.shape[0] == len(graph['nodes'])
        assert dG.TCM.shape[0] == data.shape[0]
        return 


    def test_visualize(self):
        mapper = KMapperWrapper(verbose=0, memory=None)
        data = np.random.rand(100, 3)
        graph = mapper.fit_map(data)
        y = mapper.lens_

        dG = DyNeuGraph(G=graph, y=y)

        # visualize results in a tempdir
        with tempfile.TemporaryDirectory() as temp_dir:
	        temp_html = os.path.join(temp_dir, 'test.html')
	        dG.visualize(temp_html, path_assets=temp_dir, show=False, port=None)
	        assert os.path.exists(temp_html)
        return 


    def test_visualize_show(self):
        mapper = KMapperWrapper(verbose=0, memory=None)
        data = np.random.rand(100, 3)
        graph = mapper.fit_map(data)
        y = mapper.lens_

        dG = DyNeuGraph(G=graph, y=y)

        # visualize results in a tempdir
        with tempfile.TemporaryDirectory() as temp_dir:
	        temp_html = os.path.join(temp_dir, 'test.html')
	        dG.visualize(temp_html, path_assets=temp_dir, show=False, port=None)
	        assert os.path.exists(temp_html)
        return 



""" 
Utility functions for running Mapper.
"""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals

import numpy as np
import pandas as pd

# Machine learning libraries
from sklearn import preprocessing
from sklearn import model_selection
from sklearn.cluster import DBSCAN
from sklearn.decomposition import PCA
from sklearn.manifold import TSNE, MDS 

###############################################################################
### Optional imports
###############################################################################
# External mapper tools
try:
    from kmapper import KeplerMapper
    from kmapper.cover import Cover
except ImportError as e:
    print("[warning]", e)
    

###############################################################################
### Helper functions for ptimizing various Mapper parameters
###############################################################################
def optimize_cover(X=None, r=30, g=0.67, scale_r=False, scale_g=False, ndim=2, scale_limits=False):
    """ Get optimized cover for data.

    Notes
    -----
    - Requires kmapper

    """   
    from kmapper.cover import Cover

    # Define r, g based on data / heuristic
    if X is not None:
        # Heuristic based on size, dimensionality of data
        scale_factor = (len(X) / 1000.) * (2. / ndim)
        
        # Scale r
        if scale_r:
            r = r * scale_factor

        # Scale g
        if scale_g:
            # Convert to gain, if defined as percent
            if g < 1:
                g = 1. / (1. - g)
            # Scale
            g = g * scale_factor


    # Convert gain to percent
    if g >= 1:
        g = (g - 1) / float(g)
        
    # Get n_cubes, overlap
    n_cubes = max(1, r)
    p_overlap = float(g)


    # Round final values 
    n_cubes = int(n_cubes)
    p_overlap = np.round(p_overlap, 2)

    # Define optimized limits
    limits = None
    if scale_limits is True:
        offset = p_overlap / float(n_cubes)
        limits = [[-offset, 1+offset] for _ in range(ndim)]
        n_cubes += 2 #* ndim

    try:
        # Initialize Cover with limits
        cover = Cover(n_cubes, p_overlap, limits=limits)
    except Exception as e:
        # Ignore limits, probably using older version
        cover = Cover(n_cubes, p_overlap)
        print("[warning]", e)
    return cover



def optimize_dbscan(X, k=3, p=100.0, min_samples=2, **kwargs):
    """ Get dbscan based on eps determined by data.
    """
    eps = optimize_eps(X, k=k, p=p)
    dbscan = DBSCAN(
        eps=eps, min_samples=min_samples, 
        metric='minkowski', p=2, leaf_size=15,
        **kwargs
        )
    return dbscan



def optimize_eps(X, k=3, p=100.0, **kwargs):
    """ Get optimized value for eps based on data. 

    Parameters
    ----------
    k: int
        * calculate distance to k-th nearest neighbor

    p: float 
        * threshold percentage to keep

    Returns
    -------
    eps: float
        * a parameter for DBSCAN

    """
    from sklearn.neighbors import KDTree

    # Use 'minkowski', p=2 (i.e. euclidean metric)
    tree = KDTree(X, metric='minkowski', p=2, leaf_size=15)

    # Query k nearest-neighbors for X, not including self
    dist, ind = tree.query(X, k=k+1)

    # Find eps s.t. % of points within eps of k nearest-neighbor 
    eps = np.percentile(dist[:, k], p)
    return eps



def optimize_scaler(center=True, normalize=True, **kwargs):
    """ Initialize and return a StandardScaler (default).

    Parameters
    ----------
    center: bool
        * remove the mean 

    normalize: bool
        * scale to unit variance

    Returns
    -------
    scaler: sklearn.preprocessing.StandardScaler
        * the initialized scaler

    """
    from sklearn.preprocessing import StandardScaler

    # Initialize, return the scaler
    scaler = StandardScaler(with_mean=center, with_std=normalize)
    return scaler




###############################################################################
### Helper functions for performing filtrations 
###############################################################################
def density_filter(X, k=2, inverse=True, normalize=True, **kwargs):
    """ Return function that filters the data by codensity. 

    Parameters
    ----------
    k: int
        * calculate distance to k-th nearest neighbor

    p: float 
        * threshold percentage to keep

    Returns
    -------
    indices: tuple of np.ndarrays
        * indices of core points in the data set

    """
    from sklearn.neighbors import KDTree

    # Use 'minkowski', p=2 (i.e. euclidean metric)
    tree = KDTree(X, metric='minkowski', p=2, leaf_size=15)

    # Query k nearest-neighbors for X, not including self
    dist, ind = tree.query(X, k=k+1)

    # Extract k nearest neighbors
    dens = dist[:, 1:]

    # Calculate codensity, inverse of k nearest-neighbor dists
    if inverse is True:
        dens = 1.0 / dens

    # Normalize
    if normalize is True:
        dens /= dens.max(axis=0) 

    return dens


def density_filtered_indices(X, k=15, p=90.0, **kwargs):
    """ Get sample indices based on a density filtration. 

    Parameters
    ----------
    k: int
        * calculate distance to k-th nearest neighbor

    p: float 
        * threshold percentage to keep

    Returns
    -------
    indices: tuple of np.ndarrays
        * indices of core points in the data set

    """
    from sklearn.neighbors import KDTree

    # Use 'minkowski', p=2 (i.e. euclidean metric)
    tree = KDTree(X, metric='minkowski', p=2, leaf_size=15)

    # Query k nearest-neighbors for X, not including self
    dist, ind = tree.query(X, k=k+1)

    # Find max_dist s.t. % of points within max_dist of k nearest-neighbor 
    max_dist = np.percentile(dist[:, k], p)

    # Return a mask over the data based on dist 
    indices = np.where(dist[:, k] <= max_dist)
    return indices



def random_indices(X, size=None, p=None, sort_indices=True, **kwargs):
    """ Get indices for a random subset of the data. 

    Parameters
    ----------
    size: int
        * integer size to sample (required if p=None)

    p: float 
        * threshold percentage to keep (required if size=None)

    Returns
    -------
    indices: tuple of np.ndarrays
        * indices of samples in the data set

    """
    assert(size or p)
    
    # convert p (i.e., percentage of points) to integer size
    if size is None:
        size = int(p / 100. * len(X))

    # Get original indices
    indices = np.arange(len(X))

    # Get randomized indices
    indices = np.random.choice(indices, int(size), replace=False)

    # Sort indices
    if sort_indices is True:
        indices = np.sort(indices)
    return indices



###############################################################################
### Some wrappers around the optimize helper functions
###############################################################################
def standardize_features(X, return_scaler=False, **kwargs):
    """ Standardize features in the data.
    """
    # Mean-center and normalize the data
    scaler = optimize_scaler(**kwargs)
    features = scaler.fit_transform(X)

    # Return in same format as X
    try:
        features = type(X)(features)
        features.index = X.index
        features.columns = X.columns
    except:
        pass

    # Return scaled, scaler (optional)
    if return_scaler is True:
        return features, scaler
    return features 



def filter_samples(X, method='density', return_indices=False, **kwargs):
    """ Filter core subset of samples.
    
    Parameters
    ----------
    method: str, (allowed = ['density', 'random'])
        * how to filter the samples

    """
    # Get indices based on method
    indices = np.arange(len(X))
    if 'density' in method: 
        indices = density_filtered_indices(X, **kwargs)
    elif 'random' in method:
        indices = random_indices(X, **kwargs)
    
    # Extract samples from the data
    try:
        # first, assume X is a pd.DataFrame
        samples = X.iloc[indices].copy()
    except:
        # otherwise, just return as np.ndarray
        samples = np.copy(X)[indices]

    # Return samples, indices(optional)
    if return_indices is True:
        return samples, indices
    return samples 









__all__ = [
    'graph_utils', 
    'mixture', 
    'networkx_utils', 
]

from . import graph_utils
from . import mixture
from . import networkx_utils

from .graph_utils import process_graph, extract_matrices
from .networkx_utils import format_networkx, draw_networkx
from .networkx_utils import get_cover_cubes, draw_cover
from .networkx_utils import visualize_mapper_stages






"""
Graph processing and conversion tools.
"""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals

import itertools
import functools
import json
import os

from IPython.core.display import HTML, display

from collections import defaultdict, Counter

import multiprocessing as mp

import numpy as np
import scipy as sp
import pandas as pd
from scipy import stats

from sklearn.preprocessing import Normalizer, LabelEncoder
import networkx as nx
import networkx.readwrite.json_graph


import matplotlib as mpl
import matplotlib.pyplot as plt

def in_notebook():
    """ Returns ``True`` if the module is running in IPython kernel,
      ``False`` if in IPython shell or other Python shell.
    """
    from IPython import get_ipython
    return get_ipython() is not None


def _agg_proportions(df, members=None):
    """ Aggregate proportions df for members. 
    """
    p = df.copy()
    if members is not None:
        p = p.iloc[members]
    p = p.T.assign(
        group=pd.factorize(p.columns)[0],
        label=pd.factorize(p.columns)[-1],
        value=p.sum(), #/ p.sum().sum() * p.shape[0],
        row_count=p.shape[0]
        )
    p = p[['label', 'group', 'value', 'row_count']]
    p.columns = ['label', 'group', 'value', 'row_count']
    p = list(p.T.to_dict().values())
    return p


def process_meta(meta_, labels=None, zscore=True, **kwargs):
 
    # process each column of meta
    meta_sets = dict()
    meta_labels = dict()
    for i,meta_col in enumerate(meta_.columns):
        #print("Processing meta column: {}".format(meta_col))
        meta = np.ravel(meta_[meta_col].values.copy())
        if len(meta) < 1:
            continue
        # process meta label
        meta_label = None
        if isinstance(labels, dict):
            # one list per column
            if meta_col in labels:
                meta_label = list(labels[meta_col])
        elif isinstance(labels, list):
            # shared list
            if not isinstance(labels[0], list):
                meta_label = list(labels)

        # process meta
        if str(meta[0]).isalpha() or type(meta[0]) is str:
            encoder = LabelEncoder()
            yi = encoder.fit_transform(meta)
            yi_bins = np.linspace(yi.min(), yi.max(), num=min(5, len(set(yi))), endpoint=True)
            meta = np.digitize(yi, yi_bins, right=True)
            meta = yi_bins[meta]
            meta_label = [list(yi_bins).index(_) for _ in sorted(set(meta))]
            #meta_label = ['Group '+str(_+1) for _ in meta_label]
            meta_bins = [_ for _ in zip(yi_bins[:-1], yi_bins[1:])] 
            meta_label = ['Group {} {}'.format(_+1, __) for _,__ in enumerate(meta_bins)]

        
        elif len(set(meta)) > 9 and zscore is True:
            # TODO: how do we decide whether to use continuous vs. discrete
            # zscore
            yi = meta.copy()
            yi_nz = yi[~np.isnan(yi)]
            zi = stats.zscore(yi_nz)
            zi = np.sign(zi) * np.floor(np.abs(zi))

            # digitize
            zi_bins = np.arange(zi.min(), zi.max()+1, step=1)
            zi = np.digitize(zi, zi_bins, right=True) + 1
            yi[~np.isnan(yi)] = zi
            yi[np.isnan(yi)] = 0
            yi = yi.astype(int)
            meta_bins = np.ravel([np.nan] + list(zi_bins))
            meta_label = meta_bins[np.sort(np.unique(yi))]
            meta = yi.copy()

            # map meta to labels
            #meta_str = {False: '\u03BC (= {:0.2f})'.format(np.mean(yi_nz)),
            #            True: '\u03BC + {:1.0f}\u03C3'}
            meta_str = {False: 'Mean ({:0.2f})'.format(np.mean(yi_nz)),
                        True: '{:1.0f} S.D.'}
            meta_label = [meta_str[np.abs(float(_)) > 0].format(float(_)) for _ in meta_label if not np.isnan(_)]
            
            # add n/a label if nans found
            if len(yi_nz) < len(yi):
                meta_label = ['NaN'] + meta_label
        
        elif len(set(meta)) > 9 and zscore is False:
            # TODO: figureout continuous scale here
            pass

        # labels for legend
        meta_sets[meta_col] = [_ for _ in np.sort(np.unique(meta))]
        if meta_label is not None:
            meta_labels[meta_col] = [_ for _ in meta_label]
        #print("  [+] found {} unique groups.".format(len(meta_sets[meta_col])))

        # re-assign meta
        meta_[meta_col] = meta.copy()

    return meta_, meta_sets, meta_labels


def process_graph(graph=None, meta=None, tooltips=None, color_by=None, labels=None, **kwargs):
    # convert to nx
    if isinstance(graph, nx.Graph):
        g = graph.copy()
        graph = nx.readwrite.json_graph.node_link_data(g)
    elif graph is None:
        graph = {}

    # copy graph, override defaults
    graph = dict(dict(nodes={}, links={}), **graph)

    # return empty graph
    if len(graph['nodes']) + len(graph['links']) < 1:
        return nx.Graph()

    # extract node, links
    nodelist = dict(graph.get('nodes', {}))
    edgelist = dict(graph.get('links', {}))
    memberlist = np.unique([_ for n,d in nodelist.items() for _ in d])
    nTR = int(max(np.r_[len(memberlist), memberlist+1]))
    
    # index
    node_to_index = {n:i for i,n in enumerate(nodelist)}
    index_to_node = {i:n for i,n in enumerate(nodelist)}

    # meta stuff
    if meta is None:
        meta = pd.DataFrame(index=np.arange(nTR)).assign(
            default=0,
            )
    elif not isinstance(meta, pd.DataFrame):
        meta = meta.copy()
        meta = meta.reshape(len(meta), -1)
        # check labels
        columns = ['meta-column-{}'.format(i) for i,_ in enumerate(meta.T)]
        if labels is not None and len(labels):
            columns[:len(labels)] = list(labels.keys())
        # convert to DataFrame
        meta = pd.DataFrame(
            meta, 
            columns=columns
            )
        
    # save un-processed metadata
    meta = meta.copy()
    meta_orig = meta.copy()

    # add some defaults
    meta['data_id'] = np.arange(len(meta)).astype(str)
    meta['uniform'] = '0' 


    # normalize meta
    # TODO: move all of this logic into color map utils
    #meta = Normalizer().fit_transform(meta.reshape(-1, 1))
    # bin meta (?)
    meta, meta_sets, meta_labels = process_meta(meta, labels=labels, **kwargs)
    
    # color_by
    if len(meta_orig.T) < 2:        
        color_by = color_by or meta_orig.columns[0] 

    # multiclass proportions
    # TODO: make sure this works on edge cases
    multiclass = _agg_proportions(meta_orig)
    color_by = color_by or 'multiclass' 
    meta_sets['multiclass'] = [_.get('group') for _ in multiclass]
    meta_labels['multiclass'] = [_.get('label') for _ in multiclass]
    df_multiclass = pd.DataFrame(multiclass)
    tooltip = df_multiclass.to_html(
                        index=False, columns=['label','group','value','row_count'],
                        float_format='{:0.2f}'.format,
                        )
    if kwargs.get('verbose', 1) > 0:
        if in_notebook():
            display(HTML(tooltip))
        else:
            print(df_multiclass)

    # color functions
    color_functions = kwargs.get('color_functions', {})
    #if color_by not in color_functions:
    for color_by_ in meta_sets:
        # get hex color for each group
        color_values = meta_sets[color_by_]
        n_colors = len(np.unique(color_values)) 
        if n_colors < 3:
            continue

        # check cmap       
        cmap = kwargs.get('cmap', 'tab20c')
        if not callable(cmap):
            cmap = plt.get_cmap(cmap)

        # save color values (normalized)
        if n_colors < 20:
            norm = mpl.colors.Normalize(np.min(color_values), np.max(color_values))
            color_values = norm(color_values)        

        # save color function as hex
        color_function = [mpl.colors.to_hex(_) for _ in cmap(color_values)]
        color_functions[color_by_] = color_function

    # tooltips (TODO: should this be here)
    if tooltips is None:
        tooltips = np.array([
            "<b>{}</b> (<b>size:</b> {})<br><br>".format(
                k, len(_)
            ) for k,_ in nodelist.items()
            ])
    tooltips = np.array(tooltips).astype(str)


    ### NODES
    G = nx.Graph(
        labels=meta_labels,
        groups=meta_sets,
        color_by=color_by,
        color=color_functions,
        )
    for node_id, (name, members) in enumerate(nodelist.items()):
        # define node_dict for G
        members = list(sorted(members))
        tooltip = tooltips[node_id]

        # aggregate proportions into a single column
        multiclass = _agg_proportions(meta_orig, members)
        tooltip += pd.DataFrame(multiclass).to_html(
                        index=False, header=False, columns=['label','value'],
                        float_format='{:0.2f}'.format,
                        )

        proportions = dict(
            multiclass=multiclass
            )

        # format node dict
        node_dict = dict(
            id=int(node_id),
            name=name,
            tooltip=tooltip,
            members=members,
            proportions=proportions,
            group=-1,
            # node color, size
            size=len(members),
            ) 

        coloring = kwargs.get('rsn_color','separate')
        # proportions by column
        if coloring is 'separate': # Color networks separately or put all together in one graph
            group = dict()
            for meta_col in meta.columns:
                groups = meta[meta_col].values[members] 
                proportions[meta_col] = [
                    dict(group=int(_), row_count=len(groups), value=int(c))
                    for _,c in Counter(groups).most_common() #metaset
                    ]
                group[meta_col] = int(Counter(groups).most_common()[0][0])
                # format node dict
                node_dict.update(
                    proportions=proportions, 
                    group=group
                    )  

        elif coloring is 'together':
            # multilabel
            groups = meta.values[members, :-2] 
            allgroups = [np.nonzero(memlabels) for memlabels in groups]
            allgroups = [el for subarr in allgroups for subsubarr in subarr for el in subsubarr]   
            proportions['multi'] = [
                    dict(group=int(_), row_count=len(allgroups), value=int(c))
                    for _,c in Counter(allgroups).most_common() 
                    ]
            group = int(Counter(allgroups).most_common()[0][0])
            # format node dict
            node_dict.update(
                proportions=proportions, 
                group=group
                )  
        
        # update G
        G.add_node(name, **node_dict)
        
    ### EDGES
    if not hasattr(G,'node'):
        G.node = dict(G.nodes(data=True))
        
    for (source, targets) in edgelist.items():
        # add edge for each target
        for target in targets:
            source_id = node_to_index[source]
            target_id = node_to_index[target]

            source_members = G.node[source]['members']
            target_members = G.node[target]['members']

            # add directional edges
            s_members = np.sort(source_members)
            t_members = np.sort(target_members)

            # temporal connections (i.e. (t, t+1))
            members = set(s_members).intersection(t_members)
            st_members = set(s_members + 1).intersection(t_members) - members
            ts_members = set(t_members + 1).intersection(s_members) - members

            # add forward edge
            #if len(st_members):
            #    G.add_edge(source, target, 
            #        value=len(st_members),
            #        size=len(st_members),
            #        kind='directed'
            #    )

            # add reverse edge
            #if len(ts_members):
            #    G.add_edge(target, source, 
            #        value=len(ts_members),
            #        size=len(ts_members),
            #        kind='directed'
            #    )
            # compute direction (net flow)
            n_directional = len(st_members) - len(ts_members)
            direction = np.min(t_members) - np.min(s_members)

            # define edge dict
            edge_dict = dict(
                value=len(members),
                size=len(members),
                kind='intersection',
                n_directional=n_directional
                # edge color, size
            )

            # update G
            if ((n_directional < 0) or
                (n_directional == 0 and direction < 0)):
                # net flow target to source
                G.add_edge(target, source, **edge_dict)
            else:
                # net flow source to target
                G.add_edge(source, target, **edge_dict)
                

    # degree, distance attributes (used by d3 force layout)
    for n, nbrs in G.adj.items():
        G.node[n]['degree'] = G.degree(n)
        for nbr in nbrs:
            G[n][nbr]['distance'] = 100. * (1. / min([G.degree(n), G.degree(nbr)]))

    # strength of links in the graph, based on 1 - (distance / max_distance)
    all_distances = [_ for u,v,_ in G.edges(data='distance')]
    max_distance = 1 if len(all_distances) < 1 else max(all_distances)
    for n, nbrs in G.adj.items():
        for nbr in nbrs:
            G[n][nbr]['strength'] = 1 - (G[n][nbr]['distance'] / max_distance)

    # add coloring by degree
    for n in G:
        degree = int(G.degree(n))
        size = len(G.node[n]['members'])
        G.node[n]['group']['degree'] = degree
        G.node[n]['proportions']['degree'] = [dict(
            group=degree, row_count=size, value=size
        )]

    return G





def extract_matrices(G, index=None, **kwargs):
    # construct matrices
    #   A    => adjacency matrix
    #   M    => normalized node degree
    #   T    => normalized node degree
    if index is None:
        index = np.unique([
            __ for n,_ in G.nodes(data='members') for __ in _
            ])
    nTR = int(max(np.r_[len(index), np.ravel(index)+1]))
    A = nx.to_numpy_matrix(G).A  # node x node
    M = np.zeros((nTR, A.shape[0]))    #   TR x node
    T = np.zeros((nTR, nTR))

    # return empty arrays if graph is empty
    if not len(G):
        return A, M, T

    # mapping from 'cube0_cluster0' => 0
    node_to_index = {n:i for i,n in enumerate(G)}
    node_to_members = dict(nx.get_node_attributes(G, 'members'))
    node_members = np.array(list(node_to_members.values()))

    # loop over TRs to fill in C_rc, C_tp
    for TR in range(nTR):
        # find nodes containing TR 
        TR_nodes = [n for n,_ in G.nodes(data='members') if TR in _]

        # skip if no nodes found for TR
        if not len(TR_nodes):
            if kwargs.get('verbose', 1) > 0:
                print('   > Found {} nodes for data point {}.'
                      .format(len(TR_nodes), TR))
            continue

        # display number of nodes found
        if kwargs.get('verbose', 1) > 2:
            print('   > Found {} nodes for data point {}.'
                  .format(len(TR_nodes), TR))

        # find TRs for each edge sharing node
        node_index = [node_to_index[_] for _ in TR_nodes] 
        M[TR, node_index] += 1.0

        # find similar nodes (i.e., nodes + neighbors)
        similar_nodes = np.nonzero(A[node_index,:])[-1]
        similar_nodes = np.r_[node_index,similar_nodes]
        similar_nodes = list(np.unique(similar_nodes))

        # find and count members found in similar nodes
        similar_TRs = np.hstack(node_members[similar_nodes])
        similar_TRs, counts = np.unique(similar_TRs, return_counts=True)

        # degree of connectivity b/w TRs
        T[TR, similar_TRs] += counts
    
    # symmetricize
    T = (T + T.T) / 2.0

    # normalize 
    T = T / np.max(T)    

    # return 
    return A, M, T



"""
Network plotting helper functions.
"""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals

from collections import Counter

import matplotlib as mpl
import matplotlib.pyplot as plt
import seaborn as sns

import numpy as np
import pandas as pd
import networkx as nx



###############################################################################
###  Utils for working with networkx Graphs
###############################################################################
def get_layout_pos(G, layout="spring", layout_kws={}, pos=None, lens=None, **kwargs):
    if pos is None and layout is None:
        return None

    if pos is "inverse" and lens is not None:
        pos = np.copy(lens)
        pos = {n: list(pos[_].mean(0)) for n,_ in G.nodes(data="members")}    

    if layout:
        if "spring" in str(layout) and pos is "inverse":
            layout_kws = dict(dict(iterations=1), **layout_kws)
        layout = layout if callable(layout) else getattr(nx, layout+'_layout')
        try:
            pos = layout(G, pos=pos, **layout_kws)
        except Exception as e:
            #print(e)
            pos = layout(G,  **layout_kws)
    if not isinstance(pos, dict):
        pos = {n:_ for n,_ in zip(G, pos)}
    pos = {n:list(pos[n]) for n in pos}
    nx.set_node_attributes(G, pos, "pos")
    return pos



def format_networkx(graph, meta=None, cmap="nipy_spectral_r", norm=None,  **kwargs):
    """ Format graph into networkx Graph object.
    """
    # Extract node, link data (require, for now)
    node_data = dict(graph['nodes'])
    link_dict = dict(graph['links'])
    
    # check meta data
    if meta is None:
        meta = sorted(set(__ for _ in node_data.values() for __ in _))
        
    # Construct Graph
    G = nx.Graph(link_dict)

    # color,norm
    cmap = cmap if callable(cmap) else plt.get_cmap(cmap)
    norm = norm if norm else mpl.colors.Normalize(np.min(meta)-1, np.max(meta)+1)
    to_hex = mpl.colors.to_hex
    
    # some lambdas
    _group_counter = lambda _: Counter(meta[list(_)])
    
    # Node Attributes
    nx.set_node_attributes(G, dict(node_data), "members")
    nx.set_node_attributes(G, {n: {
        "group": max(_group_counter(_)),
        "proportions": [{
            "group": g,
            "value": v,
            "row_count": len(_),
            "proportion": 100 * float(v / float(len(_))),
            } for g,v in _group_counter(_).items()],
        "row_count": len(_),
    } for n,_ in G.nodes(data="members")})


    # more lambdas
    if not hasattr(G, 'node'):
        G.node = dict(G.nodes(data=True))
    _members = lambda _: set(G.node[_].get("members"))
    _edge_members = lambda _: sorted(_members(_[0]) & _members(_[1]))
    
    # Edge Attributes
    nx.set_edge_attributes(G, {e: {
        "members": _edge_members(e),
        "group": max(_group_counter(_edge_members(e))),
    } for e in G.edges()})
    
    
    # For visualization
    nx.set_node_attributes(G, {n: {
        "color_by": norm(_.get("group")),
        "color": to_hex(cmap(norm(_.get("group")))),
        "size": len(_.get("members")),
    } for n,_ in G.nodes(data=True)})
    nx.set_edge_attributes(G, {(u,v): {
        "color_by": norm(_.get("group")),
        "color": to_hex(cmap(norm((_.get("group"))))),
        "size": len(_.get("members")),
        "weight": len(_.get("members")),
    } for u,v,_ in G.edges(data=True)})
    
    # pos
    pos = get_layout_pos(G, **kwargs)
    nx.set_node_attributes(G, pos, "pos")

            
    # Return 
    return G



def draw_networkx(graph, ax=None, fig=None, nodes=True, edges=True, **kwargs):
    """ Draw the graph as a networkx graph
    """
    import networkx as nx
    import matplotlib.pyplot as plt

    if ax is None:
        fig,ax = plt.subplots(1, 1, figsize=(5,5)) 
    else:
        fig = ax.get_figure()
    
         
    # Determine a fine size for nodes
    bbox = ax.get_window_extent().transformed(fig.dpi_scale_trans.inverted())
    width, height = bbox.width, bbox.height
    area = width * height * fig.dpi
    
    
    # format nx
    if isinstance(graph, nx.Graph):
        G = graph.copy()
    else:
        G = format_networkx(dict(graph), **kwargs)

    # pos
    pos = nx.get_node_attributes(G, "pos")
    if pos is None or not len(list(pos)):
        pos = get_layout_pos(G, **kwargs)
    kwargs.update(pos = pos)

    
    # draw edges
    if edges is True:
        edge_zorder = kwargs.pop("edge_zorder", kwargs.pop("zorder", None))
        if kwargs.get("width") is None:
            edge_w = [np.sqrt(_) for u,v,_ in G.edges(data='size')]
            kwargs.update(width=edge_w)

        if kwargs.get("edge_color") is None:
            edge_c = [_ for u,v,_ in G.edges(data="color")]
            if not any(edge_c):
                edge_c = ['grey' for _ in G.edges()]
            kwargs.update(edge_color=edge_c)

        # draw
        edges = nx.draw_networkx_edges(G, **kwargs, ax=ax)
        if edge_zorder is not None:
            edges.set_zorder(edge_zorder)

    
    # draw nodes
    if nodes is True:
        node_s0 = 0.5 * np.pi * area / G.number_of_nodes()
        node_r = np.sqrt(node_s0 / np.pi)
        node_edge = node_r / 3
        node_edge = kwargs.pop("node_edge", node_edge)
        node_edge_color = kwargs.pop("node_edge_color", "k") 
        node_zorder = kwargs.pop("node_zorder", kwargs.pop("zorder", None))

        if kwargs.get("node_size") is None:
            node_s = [node_s0 * np.sqrt(_) for n,_ in G.nodes(data="size")]
            kwargs.update(node_size=node_s)

        if kwargs.get("node_color") is None:
            node_c = [_ for n,_ in G.nodes(data="color")]
            if not any(node_c):
                node_c = [_ for n,_ in G.nodes(data="group")]
            kwargs.update(node_color=node_c)
        
        # draw
        nodes = nx.draw_networkx_nodes(G, **kwargs, ax=ax)
        if node_zorder is not None:
            nodes.set_zorder(node_zorder)
        if node_edge > 0:
            nodes.set_edgecolor(node_edge_color)
            nodes.set_linewidth(node_edge)
    
    # finish
    ax = despine(ax, **kwargs)
    return nodes, edges

def despine(ax, spines=['top','right'], **kwargs):
    if kwargs.get('keep_ticks') is not True:
        ax.set_xticks([])
        ax.set_yticks([])
    for spine in spines:
        ax.spines[spine].set_color('none')
    return ax



###############################################################################
### aliases
###############################################################################
format_nx = format_networkx
draw_nx = draw_networkx



###############################################################################
### Visualizing stages of MAPPER
###############################################################################
# TODO: move to mapper utils
def get_cover_cubes(lens=None, graph=None, cover=None, scale=False, **kwargs):
    # define bins
    cover = cover
    ids = np.arange(len(lens)).reshape(-1, 1)
    ilens = np.c_[ids, lens]

    try:
        # new Cover API from kmapper==1.2.0
        bins = np.copy(cover.centers_)
    
        # transform each node
        cover_cubes = {}
        for i, center in enumerate(cover.centers_):
            lower = center - cover.radius_
            upper = center + cover.radius_
            cover_cubes[tuple(center)] = np.r_['0,2', lower, upper]

    except Exception as e:

        # support deprecated Cover API from kmapper==1.1.6
        bins = [tuple(_) for _ in cover.define_bins(ilens)]
        chunk_dist = cover.chunk_dist
        overlap_dist = cover.overlap_dist
        d = cover.d
    
        # plot for testing
        cover_cubes = {}
        for i, cube in enumerate(bins):
            # Compute bounds
            lower = d + (cube * chunk_dist)
            upper = lower + chunk_dist + overlap_dist
            cover_cubes[tuple(cube)] = np.r_['0,2', lower, upper]

        # reindex by node
        coords = {n: tuple(graph['meta_nodes'][n]["coordinates"]) for n in graph["nodes"]}
        cover_cubes = {n: cover_cubes[coords[n]] for n in graph["nodes"]}

    if scale:
        try:
            scaler = eval(graph['meta_data']['scaler'])
        except:
            scaler = MinMaxScaler()
        # scale
        stacked = np.vstack(cover_cubes.values())
        scaler.fit(stacked)
        for cube in cover_cubes:
            _ = scaler.transform(cover_cubes[cube])
            cover_cubes[cube] = _

    return cover_cubes


def draw_cover(ax=None, cover_cubes=None, draw_all=True, max_draw=2, **kwargs):
    """ Add cover to scatter plot of projection/lens.
    """
    if ax is None:
        ax = plt.gca()

    # save for later
    xlim = np.ravel(ax.get_xlim())
    ylim = np.ravel(ax.get_ylim())
  

    if cover_cubes is None:
        cover_cubes = get_cover_cubes(**kwargs)

    # extract bins (TODO: probably a better way to do this)
    bins = np.vstack([_ for cube,_ in cover_cubes.items()])
    if len(bins.T) < 2:
        ybins = np.ravel(sorted(set(bins[:, 0])))
        bins = ybins.reshape(1, -1)
    else:
        # assume 2D
        xbins = np.ravel(sorted(set(bins[:, 0])))
        ybins = np.ravel(sorted(set(bins[:, 1])))
        bins = np.asarray([xbins, ybins])

    # save as hypercubes
    hypercubes = np.copy(bins)

    # draw
    if draw_all is True:
        max_draw = len(hypercubes)
    
    # cmap, norm for each dimension
    cmaps = [plt.get_cmap("jet")] * 2
    axspan_funcs = [ax.axhspan, ax.axvspan]
    axline_funcs = [ax.axhline, ax.axvline]
    norm = mpl.colors.Normalize(
        np.min([np.min(_) for _ in hypercubes]), 
        np.max([np.max(_) for _ in hypercubes])
        ) 

    # loop over dimensions
    for di, (axspan, axline) in enumerate(zip(axspan_funcs,axline_funcs)):
        if di >= len(hypercubes):
            continue

        # draw bounds of each cube along this dimension
        for i, hypercube in enumerate(hypercubes[di]):
            if i+1 >= len(hypercubes[di]):
                continue
            c = cmaps[di](norm(hypercubes[di][i]))
            alpha = 0.25 + (.5 * int((i+1)%2==0))
            zo = i + 1
            axspan(hypercubes[di][i], hypercubes[di][i+1], alpha=0.25*alpha, fc=c, zorder=zo)
            axline(hypercubes[di][i], alpha=alpha, c=c, zorder=zo**2)
            axline(hypercubes[di][i+1], alpha=alpha, c=c, zorder=zo**2+zo)
    
    # finish
    ax.set_xlim(xlim)
    ax.set_ylim(ylim)
    return ax

    

def visualize_mapper_stages(data, y=None, lens=None, cover=None, graph=None, dG=None, **kwargs):
    """ Visualize stages of MAPPER.

    TODO
    -----
    - this needs to be tested...
    - move to mapper utils

    """

    #### Setup
    if y is None and hasattr(data, "y"):
        y = data.y.copy()
    elif y is None and hasattr(data, "target"):
        y = data.target.copy()

    try:
        G = dG.G_.copy()
    except:
        from .graph_utils import process_graph
        G = process_graph(graph, meta=y)
    
    if not hasattr(G, 'node'):
        G.node = dict(G.nodes(data=True))

    
    # member color cmap
    cmap = kwargs.get('cmap') 
    if hasattr(data, "cmap"):
        cmap = data.cmap
    else:
        cmap = "nipy_spectral_r"
    cmap = cmap if callable(cmap) else plt.get_cmap(cmap)

    # member color norm
    norm = kwargs.get('norm') 
    if hasattr(data, "norm"):
        norm = data.norm
    else:
        norm = mpl.colors.Normalize(y.min(), y.max())
    
    # member_color 
    c = cmap(norm(y))
    c_hex = np.array([mpl.colors.to_hex(_) for _ in c])

    # node color, size
    node_size = kwargs.get('node_size')
    if node_size is None:
        node_scale_by = kwargs.get('node_scale_by', 5000)
        node_size = [node_scale_by*(len(_) / len(y)) for n,_ in G.nodes(data='members')]
    node_color = [Counter(c_hex[_]).most_common()[0][0] for n,_ in G.nodes(data='members')]


    # edge color, size
    edge_size = kwargs.get('edge_size')
    if edge_size is None:
        edge_scale_by = kwargs.get('edge_scale_by', 0.5)
        edge_size = [edge_scale_by*_ for u,v,_ in G.edges(data='size')]
    edge_color = kwargs.get('edge_color')
    if edge_color is None:
        edge_sources = [G.node[u].get('members') for u,v in G.edges()]
        edge_targets = [G.node[v].get('members') for u,v in G.edges()]
        edge_color = [Counter(c_hex[s + t]).most_common()[0][0] for s,t in zip(edge_sources, edge_targets)]


    # layout (for last stage only)
    layout = kwargs.get('layout', None)
    pos = kwargs.get('pos', None)
    if layout is None:
        pos = "inverse"

    # init figure, subplots
    figsize = kwargs.get('figsize', (20,4))
    fig, axes = plt.subplots(1, 4, figsize=figsize)

    
    #### Draw
    # ensure the lens is 2D
    lens2D = lens.copy()
    if len(lens.T) < 2:
        lens2D = np.c_[np.zeros_like(lens) + lens.mean(), lens] 
    elif len(lens.T) > 2:
        lens2D = lens2D[:, :2]

    # 1. draw lens (axes: 1-3)
    for ax in axes[:3]:
        ax.scatter(*lens2D.T, c=c, s=np.max(node_size)*.05)
        
        # adjust xlim if 1D
        if len(lens.T) < 2:
            ax.set_xlim(lens.min(), lens.max())

    # 2. draw cover (axes: 2)
    draw_cover(ax=axes[1], graph=graph, lens=lens2D, cover=cover)

    # 3. draw clusters (axes: 3)
    draw_networkx(G, lens=lens2D, pos="inverse", layout=None, 
            node_color=node_color, node_size=node_size, 
            edge_color=edge_color, width=edge_size, 
            alpha=0.5, edges=False, ax=axes[2])

    # 4. draw graph (axes: 4)
    draw_networkx(G, lens=lens2D, pos=pos, layout=layout, 
                node_color=node_color, node_size=node_size, 
                edge_color=edge_color, width=edge_size, 
                alpha=1.0, ax=axes[3])
    if layout is None:
        axes[3].set_xlim(axes[2].get_xlim())
        axes[3].set_ylim(axes[2].get_ylim())
    axes[3].axis('off')


    #### Finish
    for ax in axes:

        # despine, based on number of dimensions
        if len(lens.T) > 1:
            despine(ax, spines=['top','right'])
        else:
            despine(ax, spines=['top', 'right', 'bottom', 'left'])

        # tight layout
        ax.set_aspect('equal')
        fig.tight_layout(w_pad=2.0)

    return fig, axes







""" 
Flexible, sklearn-style wrappers for External Mapper 
"""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals

import numpy as np
import pandas as pd

# Machine learning libraries
from sklearn.datasets.base import Bunch
from sklearn.externals.joblib import Memory
from sklearn.base import BaseEstimator, TransformerMixin, ClusterMixin
from sklearn.preprocessing import MinMaxScaler
from sklearn.decomposition import PCA
from sklearn.manifold import TSNE, MDS 
from sklearn.cluster import DBSCAN

###############################################################################
### Optional imports
###############################################################################
# External mapper tools
try:
    from kmapper import KeplerMapper
    from kmapper.cover import Cover
except ImportError as e:
    print("[warning]", e)


# everything else should be in utils
from dyneusr.mapper.utils import optimize_cover, optimize_dbscan 



###############################################################################
### Helper Functions
###############################################################################
def _transform_lens(data=None, verbose=1, **params):
    """ Transform data into a lens using KeplerMapper. """
    mapper = KeplerMapper(verbose=verbose)
    return mapper.fit_transform(data, **params)


def _map_graph(lens=None, data=None, verbose=1, **params):
    """ Map lens, data into graph using KeplerMapper. """
    mapper = KeplerMapper(verbose=verbose)
    return mapper.map(lens, data, **params)

 

###############################################################################
### Base MapperWrapper
###############################################################################
class BaseMapperWrapper(BaseEstimator, TransformerMixin, ClusterMixin):

    def fit(self, data, y=None):
        pass

    def fit_transform(self, data, y=None):
        """ Transform data into lens 
        """
        return self.fit(data).lens_
    
    def fit_map(self, data, y=None):
        """ Fit lens, map data into graph.
        """
        return self.fit(data).graph_


###############################################################################
### External MapperWrappers
###############################################################################
class KMapperWrapper(BaseMapperWrapper):

    def __init__(
            self, 
            projection=None, scaler=None, 
            cover=None, clusterer=None, 
            remove_duplicate_nodes=False,
            memory='dyneusr_cache', 
            verbose=1
            ):
        """ Wraps KeplerMapper 

        Usage
        -----
            mapper = KMapperWrapper(projection=PCA(3), cover=dict(r=10, g=2))
            l = mapper.fit(X)
            g = mapper.map(l, X)

            # or 
            g = mapper.fit_map(X)
        """     
        try:
            from kmapper import KeplerMapper
            from kmapper.cover import Cover
        except ImportError as e:
            print("[warning]", e)

        # init mapper
        self.mapper = KeplerMapper()
        self.verbose = verbose

        # [1] fit params
        self.projection = projection if projection is not None else PCA(2)
        self.scaler = scaler #or MinMaxScaler()

        # [2] map params
        self.clusterer = clusterer or DBSCAN(eps=1, min_samples=2)
        self.cover = cover or Cover(10, 0.5)
        self.remove_duplicate_nodes = remove_duplicate_nodes

        # setup memory
        self.memory = Memory(memory, verbose=verbose)


    def reset(self):
        self.data_ = None
        self.lens_ = None
        self.graph_ = None
        return self
             

    def fit_lens(self, data=None, projection=None, scaler=None, **kwargs):
        """ Fit a lens over data.
        """        
        # init params
        #self.mapper = KeplerMapper(verbose=self.verbose-1)
        self.projection = projection or self.projection 
        self.scaler = scaler or self.scaler

         # fit lens
        _transform_lens_cached = self.memory.cache(_transform_lens)
        lens = _transform_lens_cached(
            data=data,  
            projection=self.projection, 
            scaler=self.scaler,
            verbose=self.verbose
            )

        # save variables
        self.data_ = data
        self.lens_ = lens
        return self



    def fit_graph(self, lens=None, data=None, clusterer=None, cover=None, **kwargs):
        """ Fit a lens over data, map data into graph.
        """
        # extract inputs
        data = self.data_ if data is None else data
        lens = self.lens_ if lens is None else lens

        # init params
        #self.mapper = KeplerMapper(verbose=self.verbose)
        self.clusterer = clusterer or self.clusterer
        self.cover = cover or self.cover

        # fit graph
        _map_graph_cached = self.memory.cache(_map_graph)
        graph = _map_graph_cached(
            lens=lens, data=data,
            clusterer=self.clusterer,
            cover=self.cover,
            verbose=self.verbose, 
            remove_duplicate_nodes=self.remove_duplicate_nodes
            )

        # save variables
        self.data_ = data
        self.lens_ = lens
        self.graph_ = graph
        return self


    def fit(self, data=None, lens=None, **kwargs):
        """ Fit a lens over data, map data into graph.
        """
        # [1] fit lens
        if lens is None:
            self.fit_lens(data, **kwargs)
            lens = self.lens_
     
        # [2] map graph
        self.fit_graph(lens, data=data, **kwargs)
        return self


###############################################################################
### wrappers as functions
###############################################################################    
def fit_kmapper(data, **params):
    mapper = KMapperWrapper(**params)
    return mapper.fit(data=data)


def run_kmapper(data, **params):
    mapper = KMapperWrapper(**params)
    mapper.fit(data=data)
    # save as bunch
    result = Bunch(
        data=mapper.data_,
        lens=mapper.lens_,
        graph=mapper.graph_,
        params=params,
        )
    return result


    
   
   


"""
Trefoil knot data loader
"""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals


import logging
logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)

from itertools import product 

import numpy as np
import pandas as pd 

from sklearn.datasets.base import Bunch

import matplotlib as mpl
#mpl.use('TkAgg', warn=False)
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D

import seaborn as sns
sns.set("paper", "white")


##############################################################################
### data generators
##############################################################################
def make_trefoil(size=100, noise=0.0, a=2, b=2, c=3, **kwargs):
    """Generate synthetic trefoil dataset.

    Params
    -----
    :size = int (default = 1000)
     - the number of data points to use
    """
    # generate trefoil
    phi = np.linspace(0, 2*np.pi, size)
    x = np.sin(phi) + a*np.sin(a*phi)
    y = np.cos(phi) - b*np.cos(b*phi)
    z = -np.sin(c*phi)

    # add noise
    if noise is not None and noise > 0.0:
        noise = np.ravel(noise)
        if noise.shape[0] == 1:
            noise = noise.repeat(3)
        x += np.random.normal(0, noise[0], size)
        y += np.random.normal(0, noise[1], size)
        z += np.random.normal(0, noise[2], size)
    
    # remove tiny numbers
    x[np.abs(x) < 1e-6] = 0
    y[np.abs(y) < 1e-6] = 0
    z[np.abs(z) < 1e-6] = 0

    # stack features
    data = np.c_[x, y, z][:]
    data_ids = np.arange(phi.shape[0])

    # define target
    target = np.copy(phi)
    target = np.roll(target, int(size*0.05))

    # coloring based on target
    cmap = plt.get_cmap('brg', 3)
    norm = mpl.colors.Normalize(target.min(), target.max())
    coloring = cmap(norm(target))


    # format data bunch
    dataset = Bunch(
        data=data,
        feature_names=["x", "y", "z"],
        target=target,
        coloring=coloring,
        cmap=cmap, norm=norm,
        # other meta-data
        index=data_ids,
        domain=phi,
        )
    return dataset



##############################################################################
### data visualizers
##############################################################################
def draw_trefoil3d(x=None, y=None, z=None, c=None, s='z', ax=None, fig=None, view=(90, -90), **kwargs):
    """Plot trefoil knot.
    """   
    if c is None:
    	c = np.copy(z)
    cmap = plt.get_cmap(kwargs.get('cmap', "brg"), 3)
    norm = mpl.colors.Normalize(c.min(), c.max())

    # extract x, y, z
    if s == 'z':
        zbins = np.linspace(z.min(), z.max(), num=10)
        zbins = np.digitize(z, zbins) 
        s = zbins**2

    # combine features
    X = np.c_[x, y, z]

    # plot data
    fig, axes = plt.subplots(1, 3, figsize=(15,5),subplot_kw=dict(projection='3d'))

    # 3 views
    for ax_i, ax in enumerate(axes):

        if ax_i == 0:
            xcol, ycol, zcol = 0, 1, 2
        elif ax_i == 1:
            xcol, ycol, zcol = 0, 2, 1
        elif ax_i == 2:
            xcol, ycol, zcol = 1, 2, 0
        zbins = np.linspace(z.min(), z.max(), num=10)
        zbins = np.digitize(z, zbins) 
        ax.scatter(X[:,xcol], X[:,ycol], X[:,zcol], c=c, s=s, alpha=0.8, cmap=cmap, norm=norm)
        ax.set_xlabel(list('xyz')[xcol], fontweight='bold')
        ax.set_ylabel(list('xyz')[ycol], fontweight='bold')
        ax.set_zlabel(list('xyz')[zcol], fontweight='bold')
        ax.set_xticks([])
        ax.set_yticks([])
        ax.set_zticks([])

        # Get rid of colored axes planes
        # First remove fill
        ax.xaxis.pane.fill = False
        ax.yaxis.pane.fill = False
        ax.zaxis.pane.fill = False

        # Now set color to white (or whatever is "invisible")
        ax.xaxis.pane.set_edgecolor('w')
        ax.yaxis.pane.set_edgecolor('w')
        ax.zaxis.pane.set_edgecolor('w')

        # Bonus: To get rid of the grid as well:
        ax.grid(False)
        #if ax_i == 1:
        #    view = (view[0]+0, view[1]+90)
        #if ax_i == 2:
        #    view = (view[0]+90, view[1]+0)

        ax.set_title("view = {}".format(view))
        ax.view_init(*view)

    return axes



def draw_trefoil(x=None, y=None, z=None, c=None, s='z', ax=None, fig=None, **kwargs):
    """Plot trefoil knot.
    """   
    if c is None:
        c = np.copy(z)
    cmap = plt.get_cmap(kwargs.get('cmap', "brg"), 3)
    norm = mpl.colors.Normalize(c.min(), c.max())

    # extract x, y, z
    if s == 'z':
        zbins = np.linspace(z.min(), z.max(), num=10)
        zbins = np.digitize(z, zbins) 
        s = zbins**2

    # combine features
    X = np.c_[x, y, z]

    # plot data
    fig, axes = plt.subplots(3, 1, figsize=(15,5))

    # subplots of each dim as time-series
    for ax_i, (col_name, col_ys) in enumerate(dict(x=x, y=y, z=z).items()):
        ax = axes[ax_i]
        ax.scatter(np.arange(len(col_ys)), col_ys, c=c, s=s, cmap=cmap, norm=norm)
        ax.set_ylabel(col_name, fontweight='bold')
        if ax_i < 2:
            ax.set_xticks([])
        else:
            ax.set_xlabel('index', fontweight='bold')
    return axes






"""
D3.js visualizations and helper functions.
"""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals

import os
import json 
import subprocess
import functools 

from pathlib import Path

import IPython
import matplotlib.pyplot as plt
import numpy as np



def format_IFrame(path, **kwargs):
    defaults = dict(src=path, width="100%", height=800, frameBorder=0)
    return IPython.display.IFrame(**dict(defaults, **kwargs))


CUSTOM_CSS = """
<style>
  .container { width:80% !important; margin:auto }
  .output_scroll {height: 80px !important;}
</style>
"""

def display_HTML(html="", src="", figure=None, static=False):
    """ Display html inside a Jupyter notebook.

    Reference
    ---------
    - [1] https://github.com/MLWave/kepler-mapper/issues/10
    """
    # check figure
    figure = figure or plt.figure()
    # format CSS, HTML (use IFrame if static=False)
    css = IPython.display.HTML(CUSTOM_CSS)
    html = IPython.display.HTML(html) if static else format_IFrame(src)
    # display
    IPython.display.display(css)
    IPython.display.display(html)
    return html

def in_notebook():
    """ Returns ``True`` if the module is running in IPython kernel,
      ``False`` if in IPython shell or other Python shell.
    """
    from IPython import get_ipython
    return get_ipython() is not None

def json_dump(obj, fp):
    """ Handles numpy data types.

    Reference
    ---------
    - [1] https://stackoverflow.com/questions/26646362/
    """
    def default(o):
        if type(o).__module__ == np.__name__:
            if isinstance(o, np.ndarray):
                return o.tolist()
            else:
                return o.item()
        return json.JSONEncoder().default(o)

    json.dump(obj, fp, default=default)
    return None




def http_server(port=None, host='localhost'):
    """ Start http.server listening to port.

    Usage
    -----
    # open specific port
    status = open_port(8800)

    # open next available
    port = open_port(None)

    """     
    if port is not None and port < 0:
        # open http.server on port, or next available
        p = http_server(port=8000, host=host) 

        # check status, try next (until found)
        while p.status is False:
            p = http_server(port=p.port+1, host=host)
        return p

    else:
        # open http.server on port
        p = subprocess.Popen(
            'python3 -m http.server {} -b {} &'.format(port, host), 
            shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE
            )
        p.host = host
        p.port = port


        # check status, return process
        try:
            o, e = p.communicate(timeout=1)
            print("Already serving {}:{} ".format(host, port))
            p.status = False
        except subprocess.TimeoutExpired as e:
            print("Serving HTTP on {}:{} ...".format(host, port))
            p.status = True        
        return p
                


    
def visualize_force(js, template=None, path_html='index.html', path_csv=None, path_json=None, path_graphs='graphs', path_assets=None, reset=True, static=False, show=True, figure=None, port=8000, **kwargs):
    """ Create index.html, index.csv, graphs/*.json
    """
    ### check for deprecated options
    if kwargs.get('PORT'):
        # PORT => port
        port = kwargs.pop('PORT')

    ### no need to open a new port if static
    if static is True:
        port = None
        
    ### Run http.server on port
    HTTP = http_server(port=port, **kwargs)

    ### Read template HTML
    file_template = 'index.html'
    if template and template not in ('movie',): 
        file_template = file_template.replace('.html', '-{}.html'.format(template))
    path_template = Path(__file__).resolve().parents[0] / 'templates' / file_template
    with path_template.open('r') as f:
        html = f.read()

    ### Define path to output HTML
    path_html = Path(path_html)
    url = 'http://{}:{}/{}'.format(HTTP.host, HTTP.port, str(path_html))

    ### Path to save assests
    if path_assets is None:
        path_assets = '.'
    path_assets = Path(path_assets)
    if not path_assets.exists():
        os.makedirs(str(path_assets))

    ### Check graphs director
    graphs_dir = path_html.parents[0] / path_assets / path_graphs
    if not graphs_dir.exists():
        os.makedirs(str(graphs_dir))

    ### Write graphs/*.json
    if path_json is None:
        path_json = graphs_dir / path_html.name.replace('.html', '.json')
    path_json = Path(path_json)
    with path_json.open('w') as f:
        json_dump(js, f)
        

    ### Write index.csv
    if path_csv is None:
        #path_csv =  path_html.parents[0] / path_assets / path_html.name.replace('.html', '.csv')
        path_csv = graphs_dir / path_html.name.replace('.html', '.csv')
    # clean up paths
    path_csv = Path(path_csv)
    path_csv_rel = path_csv.relative_to(path_html.parents[0])
    path_json_rel = path_json.relative_to(path_html.parents[0])
    # update html with csv path
    html = html.replace('index.csv', str(path_csv_rel))
    if path_csv.exists() and not reset:
        with path_csv.open('a') as f:
            f.write('\n')
            f.write(str(path_json_rel))
    else:
        with path_csv.open('w') as f:
            f.write('json\n')
            f.write(str(path_json_rel))

    ### Load template HTML
    # TODO: should probably seperate these into seperate files
    if static is True:        
        ### Rename path with -static
        #path_html = Path(str(path_html).replace('.html', "-static.html"))
        url = path_html.absolute().as_uri()

        ### Add js directly to HTML
        with path_json.open('r') as f:
            json_str = json.load(f)
            json_str = json.dumps(json_str).replace("\"", "\"")
            div = """\n\t<div id="json_graph" data-json='{json_graph}' style="display:none"></div>"""
            div = div.format(json_graph=json_str)    
            html = html.replace('</head>', div+'\n\n</head>')


    ### Write HTML to file
    print('[Force Graph] {}'.format(str(url)))
    with path_html.open('w') as f:
        f.write(html)    

    ### Show HTML
    if show and in_notebook():
        display_HTML(html=html, src=url, figure=figure, static=static)
    elif show is True:
        try:
            import webbrowser
            webbrowser.open(str(url))
        except ImportError as e:
            print(e)
            print('Hint: requires Python webbrowser module...')

    # cache some things
    HTTP.url = str(url)
    HTTP.src = str(path_html)
    HTTP.html = str(html)
    HTTP.json = dict(js)

    # define some helper functions (TODO: clean up HTTP handling, move to utils)
    HTTP.display = functools.partial(display_HTML, html=HTTP.html, src=HTTP.url, static=static)
    try:
        import webbrowser
        HTTP.open = functools.partial(webbrowser.open, url=HTTP.url)
    except:
        HTTP.open = functools.partial(print, HTTP.url)

    return HTTP



import sys
import warnings

if not sys.warnoptions:
    warnings.simplefilter("ignore")


__all__ = [
    'datasets',
    'mapper',
    'tools',
    'visuals',
    'core',
]

from . import datasets
from . import mapper
from . import tools
from . import visuals
from . import core

from .core import DyNeuGraph
from ._version import __version__



__all__ = [
    'utils',
    'wrappers',
    #'persistence' 
]

from . import utils
from . import wrappers
# from . import persistence

from .utils import optimize_cover, optimize_eps, optimize_dbscan
from .wrappers import KMapperWrapper, run_kmapper




"""
Mixture model tools.
"""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals

import os
import numpy as np
import scipy.stats

from sklearn.preprocessing import StandardScaler, Normalizer

from nilearn.input_data import NiftiMapsMasker, NiftiLabelsMasker
from nilearn.datasets import fetch_atlas_msdl
from nilearn.connectome import ConnectivityMeasure
from nilearn import image
from nilearn import plotting

import matplotlib as mpl
import matplotlib.pyplot as plt



######################################################################
###
######################################################################
def mean_mixture(data, index=None, fwhm=None, threshold=None, **kwargs):
    # extract array
    X = data
    if hasattr(X, 'X'):
    	X = data.X
    if hasattr(X, 'data'):
    	X = X.data
    if hasattr(X, 'values'):
    	X = X.values

    # extract rows based on index
    mm_X = X[list(index), :].copy() # 
    mm_X = mm_X.mean(axis=0, keepdims=True)

    # unmask
    mm_img = data.masker.inverse_transform(mm_X)
  
    # smooth, threshold
    #mm_img = image.mean_img(mm_img)
    if fwhm is not None:
        mm_img = image.smooth_img(mm_img, fwhm=fwhm)        
    if threshold is not None:
        mm_img = image.threshold_img(mm_img, threshold)
    return mm_img


def simple_mixture(data, index=None, agg='mean', fwhm=None, threshold=None, **kwargs):
    # extract array
    X = data
    if hasattr(X, 'X'):
    	X = data.X
    if hasattr(X, 'data'):
    	X = X.data
    if hasattr(X, 'values'):
    	X = X.values

    # extract rows based on index
    mm_X = X[list(index), :].copy()
    mm_pos_X = np.copy(mm_X)
    mm_neg_X = np.copy(mm_X)

    # zero out pos, neg in opposite array
    mm_pos_X[(mm_X < 0.0)] = 0.0
    mm_neg_X[(mm_X > 0.0)] = 0.0

    # means
    # TODO: use StandardScaler().fit_transform(...)
    mean_pos_X = mm_pos_X.mean(axis=0) / (mm_pos_X.std(axis=0)+1) * len(index)
    mean_neg_X = mm_neg_X.mean(axis=0) / (mm_neg_X.std(axis=0)+1) * len(index)
    stack_pos_neg_X = np.stack([mean_pos_X, mean_neg_X])

    # agg over stack
    np_agg = agg if callable(agg) else getattr(np, agg)
    pos_neg_X = np_agg(stack_pos_neg_X, axis=0, keepdims=True)
    mm_X = mm_X.mean(axis=0, keepdims=True)#pos_neg_X.astype(np.float32)

    # unmask
    #mm_X = scipy.stats.zscore(mm_X, axis=-1)
    mm_img = data.masker.inverse_transform(mm_X)

    # clean
    #mm_img = image.clean_img(mm_img, standardize=True)

    # smooth, threshold
    if fwhm is not None:
        mm_img = image.smooth_img(mm_img, fwhm=fwhm)        
    if threshold is not None:
        mm_img = image.threshold_img(mm_img, threshold)
    return mm_img


def simple_mixtures(data, mixtures=[], targets=None, kind='simple', mode='glass', prefix='timestep', save_dir='tooltips', show_every_n=0, print_every_n=1, figsize=(4,4), plot_kws=dict(), **kwargs):
    # make sure output path exists
    if not os.path.exists(save_dir):
        os.makedirs(save_dir)

    # check mixtures
    if isinstance(mixtures, dict):
        mixtures = mixtures.items()
    else:
        mixtures = list(enumerate(mixtures))
   
    # loop over mixtures
    filenames = []    
    for i, (id_, mixture) in enumerate(mixtures):
        # continue
        if not len(mixture):
            continue

        # format path to save figure
        save_as = os.path.join(save_dir, '{}_MM_{}{}.jpg'.format(kind, prefix, id_))
        filenames.append(save_as)  

        # title
        title = "{}{}".format(prefix, id_)
        try:
            target = targets[id_]
            title += " ({})".format(target)
        except:
            pass

    
        # plot, if file does not already exists
        if os.path.exists(save_as):

            # show the image, even if it already exists
            if show_every_n>0 and (i)%show_every_n == 0:
                plt.figure(figsize=figsize)
                plt.imshow(plt.imread(save_as))
                plt.axis('off')
                plt.show()
            continue

        # get simple mixture model
        mm = None
        if kind == 'mean':
            mm = mean_mixture(data, index=mixture, **kwargs)
        else:
            mm = simple_mixture(data, index=mixture, **kwargs)


        # plot simple mixture
        fig = plt.figure(figsize=figsize)
        if 'glass' in mode:
            # setup for plotting
            plot_kws = dict(dict(
                display_mode='z', plot_abs=False, 
                vmin=-3, vmax=3, #threshold=0.5,
                #cmap='jet', 
                colorbar=True,
                ), **plot_kws)
            # plot
            plotting.plot_glass_brain(mm, figure=fig, title=title, **plot_kws)
        else:
            # setup for plotting
            plot_kws = dict(dict(
                display_mode='z', cut_coords=1, 
                #threshold=0.5, 
                ), **plot_kws)
            # plot
            plotting.plot_stat_map(mm, figure=fig, title=title, **plot_kws)


        # save, show, close
        plt.savefig(save_as, dpi=300, transparent=True)
        if show_every_n>0 and (i)%show_every_n == 0:
            plt.show()
        plt.close('all')
      
        # display progress
        if print_every_n>0 and (i)%print_every_n == 0:
            print("[{} of {}] Saved: {}".format(i, len(mixtures), save_as))
    print('[done]')
    return filenames



######################################################################
###
######################################################################
def connectome_mixtures(data, mixtures=[], metric="correlation", prefix='', save_dir='figures', show=False, reset=False, **plot_kws):
    filenames = []
    if isinstance(mixtures, dict):
        mixtures = mixtures.items()
    else:
        mixtures = enumerate(mixtures)
    for id_, mixture in mixtures:
        print("Building Mixture Connectome... ID: {} (size = {})".format(id_, len(mixture)))
        save_as = 'CC_{}__{}{}.png'.format(metric, prefix, id_).replace(' ','_')
        save_as = os.path.join(save_dir, save_as)
        if not os.path.exists(os.path.dirname(save_as)):
            os.makedirs(os.path.dirname(save_as))
        if not os.path.exists(save_as) or reset is True:
            display = plot_connectome_mixture(data, index=mixture, metric=metric, save_as=save_as, show=show, **plot_kws)          
        # save filename
        filenames.append(save_as)
    return filenames



def plot_connectome_mixture(data=None, index=None, metric="correlation", save_as=None, show=False, **kwargs):
   
    # extract time series from all subjects and concatenate them
    mm = data.X[index, :].copy()
    time_series = [np.vstack(mm)]

    # calculate correlation matrices across indexed frames in data 
    connectome_measure = ConnectivityMeasure(kind=metric)
    connectome_measure.fit_transform(time_series)
    connectivity = connectome_measure.mean_
    np.fill_diagonal(connectivity, 0)
    #connectivity[np.abs(connectivity) < 0.2] = 0.0 


    # grab center coordinates for atlas labels
    atlas = kwargs.pop('atlas', data.atlas)
    coords = plotting.find_parcellation_cut_coords(labels_img=atlas)

    # assign node colors
    cmap = kwargs.pop('cmap', 'jet')
    node_cmap = plt.get_cmap('bone')
    node_norm = mpl.colors.Normalize(vmin=-0.8, vmax=1.2)
    node_colors = np.ravel([_[-1] for i,_ in enumerate(coords)])
    node_colors = [_ / np.max(node_colors) for _ in node_colors]
    node_colors = node_cmap(node_norm(node_colors))

    # plot connectome matrix
    fig = plt.figure(figsize=(12,5))
    ax = plt.subplot2grid((1, 2), (0, 1),  rowspan=1, colspan=1) 
    display = plotting.plot_matrix(
        connectivity,
        vmin=-.5, vmax=.5, colorbar=True, cmap=cmap,
        axes=ax, #title='{} Matrix'.format(metric.title()),
        )

    # plot connectome with 99.7% edge strength in the connectivity
    ax = plt.subplot2grid((1, 2), (0, 0), rowspan=1, colspan=1)
    display = plotting.plot_connectome(
        connectivity, coords,
        edge_threshold="99.9%", display_mode='z',
        node_color=node_colors, node_size=20, edge_kwargs=dict(lw=4),
        edge_vmin=-.8, edge_vmax=.8, edge_cmap=cmap,
        colorbar=False, black_bg=not True, alpha=0.5,
        annotate=False,
        axes=ax,
        )
    if show is True:
        plt.show()
    plt.subplots_adjust(left=0.05, right=0.95, bottom=0.05, top=0.95)
    if save_as:
        fig.savefig(save_as, transparent=True)#, facecolor='slategray', edgecolor='white')
    plt.close(fig)
    return display



def spatial_correlations(mm, atlas=None):
    if atlas is None:
        from nilearn.datasets import fetch_atlas_msdl
        atlas = fetch_atlas_msdl()
    from nilearn.input_data import NiftiMapsMasker
    masker = NiftiMapsMasker(maps_img=atlas['maps'])
    rsns_masked = masker.fit_transform(atlas['maps'])
    mm_masked = masker.fit_transform([mm])
    cc = np.corrcoef(mm_masked, rsns_masked)
    return cc


def run_jobs(jobs, multiproc=False):
    # 3. get results
    results = dict()
    if multiproc:     
        print("Running jobs... (multiproc={})".format(multiproc))
        import multiprocessing as mp
        pool = mp.Pool(mp.cpu_count()-1)
        results = {_:pool.apply_async(eval, jobs[_]) for _ in jobs}
        results = {_:results[_].get() for _ in results}
        pool.close()
        pool.join()
    else:
        print("Running jobs... (multiproc={})".format(multiproc))
        for i, job in jobs.items():
            print("[job, i={}] {}".format(i, job))
            results[tr] = eval(job)
    # return
    return results



"""
Graph annotation helper functions.
"""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals

import os
import json

from pathlib import Path
import matplotlib.pyplot as plt

import numpy as np
import networkx as nx



NODE_HTML = """
<div id="tooltip-content">
  <div><b>Node ID:</b> {node_name}</div>
  <div><b>Node Size:</b> {node_size}</div>
  <div><b>Members:</b></div>
  <br>

  {member_tooltips}

</div>
"""

MEMBER_HTML = """
<span>{member_id} </span>
"""

IMG_HTML = """
<img class="tooltip-image" 
     src="{src}" 
     style="box-shadow: 0px 0px 0px 0px {color}">
"""


def annotate(G, verbose=False, **kwargs):
    """ Annotate G, format tooltips.
    
        #kwargs['image'][member_ids]
    """
    # node data = [(n, {...}), ...]
    node_data = {}
    if isinstance(G, nx.Graph):
        node_data = G.nodes(True)
    elif 'nodes' in G:
        node_data = [(d["id"], d) for d in G['nodes']]
    
    # unique ids
    member_uids = np.sort(np.unique([
        __ for n,d in node_data for __ in d['members']
        ]))

    # annotations
    annotation_map = dict(G.nodes(data=True))
    for ni,(n,d) in enumerate(node_data):
        
        # progress
        if verbose:
            print("Annotating node... NodeID:", n)
        member_ids = d['members']
        
        # member_html
        members_html = ""
        if kwargs.get('image') is not None:
            images = np.ravel(kwargs.get('image', []))
            if len(images) == 1:
                images = images[[0]]
            elif len(images) < len(member_uids):
                images = images[[ni]]
            else:
                images = images[member_ids]
            images = [IMG_HTML.format(src=_, color="#ffffff") for _ in images]
            members_html += "".join(images)
        else:
            members_htmls = [MEMBER_HTML.format(
                member_id=member_id,
                member_color=dict(node_data).get("color", "#777")
            ) for member_id in member_ids]  
            members_html += "".join(members_htmls)          

        # format tooltip for node
        node_tooltip = annotation_map[n].get('tooltip', '')
        if len(node_tooltip) < len(NODE_HTML):# is None: 
            node_tooltip = NODE_HTML.format(
                node_name=d['name'],
                node_size=len(member_ids), 
                member_tooltips=members_html
            )
        else:
            node_tooltip += """<div>{}</div>""".format(members_html)
    
        # annote node
        annotation_map[n] = dict(
            tooltip=node_tooltip, 
            image=members_html,
            )

    # nx
    if verbose:
        print("Annotating Graph...")
    nx.set_node_attributes(G, annotation_map)
    return G, annotation_map



def format_tooltips(G, **kwargs):
    """ Annotate G, format tooltips.
    """
    # node data = [(n, {...}), ...]
    node_data = {}
    if isinstance(G, nx.Graph):
        node_data = G.nodes(True)
    elif 'nodes' in G:
        node_data = [(d["id"], d) for d in G['nodes']]

    # unique ids
    member_uids = np.sort(np.unique([
        __ for n,d in node_data for __ in d['members']
        ]))

    # array of tooltips
    node_tooltips = []
    for n,d in node_data:

        # progress
        print("Formatting tooltip... NodeID:", n)
        member_ids = d['members']

        # member images
        images = d['image'][member_ids]
        images = [IMG_HTML.format(src=_) for _ in images]
    
        # format tooltip for node
        node_tooltip = NODE_HTML.format(
            node_id=n, node_name=d['name'],
            node_size=len(member_ids), 
            data_size=len(member_uids),
            images=images
        )
    
        # add to array
        node_tooltips.append(node_tooltip)
    
    # make numpy array
    return np.array(node_tooltips)



"""
Plotting wrappers and helper functions.
"""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals

import os

import numpy as np
import pandas as pd

import matplotlib as mpl
import matplotlib.pyplot as plt


import seaborn as sns 

from sklearn.preprocessing import Normalizer, Binarizer



def plot_temporal_degree(TCM, y=None, save_as=None, show=False, scalers=[Normalizer()], cmap='plasma', windows=None, **kwargs):
        """ Plot temporal transitions

        """
        def draw_axlines(*axline_funcs):
            """ Draw axlines based on y groups
            """
            if y is None:
                return 
            for i, y_ in enumerate(y):
                if i+1==y.shape[0] or y_ == y[i+1]:
                    continue
                for axline_func in axline_funcs:
                    axline_func(i, color="darkslategray")
            return

        # copy tcm
        tcm = np.copy(TCM)

        # scale ?
        #for scaler in filter(None, np.ravel(scalers)):
        #    tcm = scaler.fit_transform(tcm)
    
        figs = dict()

        ### heatmap of TCM      
        color = kwargs.get('color', 'steelblue')

        # setup figure, unless already exists
        fig = kwargs.get('fig')
        ax = kwargs.get('ax')
        if ax is not None:
            fig = ax.get_figure()
        elif fig is not None:
            ax = plt.gca()
        else:
            fig = plt.figure(figsize=(20,5))
            ax = plt.subplot2grid((5, 5), (0, 0), rowspan=5, colspan=5)

        ### tsplot
        #fig = plt.figure(figsize=(5,2))
        draw_axlines(plt.axvline)

        # normalized degree
        #deg = (tcm > 0.0).astype(float).sum(axis=1)
        deg = tcm.sum(axis=1)
        deg /= deg.max()
        
        # moving average
        window = kwargs.pop('window', 1) 
        deg_mva = deg.copy()
        if window > 1:
            rolling_kind = kwargs.get('rolling', np.mean)
            deg_rolling = pd.DataFrame(deg).rolling(window, center=True, min_periods=int(window/2))
            deg_mva = deg_rolling.apply(rolling_kind, raw=True)
            deg_mva /= deg_mva.max()
        # scale ?
        #deg_mva = deg_mva.reshape(-1, 1)
        #for scaler in filter(None, np.ravel(scalers)):
        #    deg_mva = scaler.fit_transform(deg_mva)

   
        # plot
        trs = np.arange(deg.shape[0])
        plt.plot(trs, np.ravel(deg),ls='-', marker='', alpha=0.3, color=color)
        plt.plot(trs, np.ravel(deg_mva)[:len(trs)], alpha=1.0,lw=3, color=color)

        ax = plt.gca()
        #ax.set_ylim(0.0, 1.1)
        xlim = kwargs.get('xlim', (trs.min(), trs.max()))
        ax.set_xlim(xlim)
        
        if xlim is not None:
            ax.set_xlim(xlim)

        title = kwargs.get('title') or 'Degree (TCM)'
        if title:
            ax.set_xlabel("Time frame (TR)", fontweight='bold', fontsize=16)
            ax.set_ylabel("Normalized degree", fontweight='bold', fontsize=16)
            ax.set_title(title, fontweight='bold', fontsize=18)
        fig.tight_layout(rect=(0.0, 0.03, 1.0, 0.98))
        #figs['tsplot'] = fig
        if save_as is not None:
            print("Saving figure, save_as:", save_as)
            if not os.path.exists(os.path.dirname(save_as)):
                os.makedirs(os.path.dirname(save_as))
            fig.savefig(save_as, transparent=True)#, facecolor='lightgray', edgecolor='w')

        # show
        if show is True:
            plt.show()
        
        # save
        return fig, ax



# -----------------------------------------------------------
def plot_scatter(x, y, meta=None, fig=None, **scatter_kw):
    
    # figure
    fig = fig or plt.figure(figsize=(15,3))
    
    # update scatter_kw with colors
    if scatter_kw.get('c') is None:
        scatter_kw['c'] = meta.colors
    
    # take a look at the projection
    ax = fig.add_subplot(141)
    ax.scatter(x, y, **scatter_kw)
    ax.set_xlabel('Lens 1', fontweight='bold')
    ax.set_ylabel('Lens 2', fontweight='bold')
    
    # lets try again, this time in 3d (over time)
    from mpl_toolkits.mplot3d import Axes3D
    z = range(len(x))
    
    # second view
    ax = fig.add_subplot(142, projection='3d')
    ax.scatter(z, x, y, **scatter_kw)
    ax.set_xlabel('TR', fontweight='bold')
    ax.set_ylabel('Lens 1', fontweight='bold')
    ax.set_zlabel('Lens 2', fontweight='bold')
    ax.view_init(5, 0)
    
    # second view
    ax = fig.add_subplot(143, projection='3d')
    ax.scatter(y, z, x, **scatter_kw)
    ax.set_xlabel('Lens 2', fontweight='bold')
    ax.set_ylabel('TR', fontweight='bold')
    ax.set_zlabel('Lens 1', fontweight='bold')
    ax.view_init(5, 0)
    
    # first view
    ax = fig.add_subplot(144, projection='3d')
    ax.scatter(x, z, y, **scatter_kw)
    ax.set_xlabel('Lens 1', fontweight='bold')
    ax.set_ylabel('TR', fontweight='bold')
    ax.set_zlabel('Lens 2', fontweight='bold')
    ax.view_init(5, 0)

    return fig


def plot_multiscatter(x, y, meta=None, **scatter_kw):
    figs = []
    for group_i, group in enumerate(sorted(meta.groups.unique())):
        group_meta = meta.loc[meta.groups.eq(group), :]
        fig = plot_scatter(
            x[group_meta.index.values], y[group_meta.index.values],
            meta=group_meta,
        )
        plt.title('Group {}'.format(group))
        figs.append(fig)
    return figs



from setuptools import find_packages, setup
import re

# parse dyneusr/_version.py
try:
    version_fn = 'dyneusr/_version.py'
    with open(version_fn) as version_fd:
        version = version_fd.read()
    version_re = r"^__version__ = ['\"]([^'\"]*)['\"]"
    version = re.findall(version_re, version, re.M)[0]
except:
    raise RuntimeError("Unable to read version in {}.".format(version_fn))


# parse requirements.txt
with open('requirements.txt') as f:
    install_requires = [_ for _ in f.read().split('\n') 
                        if len(_) and _[0].isalpha()]

# parse README.md
with open('README.md') as f:
    long_description = f.read()

# run setup
setup(
    name='dyneusr',
    version=version,
    description='Dynamical Neural Spatiotemporal Representations.',
    long_description=long_description,
    long_description_content_type="text/markdown",	
    author='Caleb Geniesse',
    author_email='geniesse@stanford.edu',
    url='https://braindynamicslab.github.io/dyneusr',
    license='BSD-3',
    packages=find_packages(),
    include_package_data=True,
    install_requires=install_requires,
    python_requires='>=3.6',
    classifiers=[
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Information Analysis",
        "Topic :: Scientific/Engineering :: Visualization",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
    ],
    keywords='brain dynamics, topology data analysis, neuroimaging, brain networks, mapper, visualization',
)



__all__ = [
    'annotation',
    'plotting', 
    'visualize', 
]

from . import annotation
from . import plotting
from . import visualize 

from .annotation import annotate, format_tooltips
from .plotting import plot_temporal_degree
from .visualize import visualize_force



import pytest
import numpy as np
np.random.seed(123)

import dyneusr as ds
from dyneusr.mapper.wrappers import KMapperWrapper
from dyneusr.mapper.utils import density_filter


class TestWrappers:
    """ Simple tests for mapper.wrappers.py
    """

    def test_KMapperWrapper(self):
        mapper = KMapperWrapper(verbose=0, memory=None)
        data = np.random.rand(100, 3)
        graph = mapper.fit_map(data)
        assert mapper.lens_.shape[0] == data.shape[0]
        assert graph.get('nodes') and graph.get('links')
        return 




class TestUtils:
    """ Simple tests for mapper.utils.py
    """

    def test_density_filter(self):
        mapper = KMapperWrapper(verbose=0, memory=None)
        data_a = np.random.rand(100, 3) 
        data_b = np.random.rand(100, 3) * 10

        lens_a = density_filter(data_a, k=3)
        lens_b = density_filter(data_b, k=3)

        assert data_a.shape[0] == lens_a.shape[0]
        assert data_b.shape[0] == lens_b.shape[0]
        assert lens_a.shape[1] == 3
        assert lens_b.shape[1] == 3

        # make sure that b has lower lowest density
        assert lens_a.min() > lens_b.min()
        return 





"""
DyNeuGraph class definition
"""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals

import os
import json

import numpy as np 
import pandas as pd
import networkx as nx
import networkx.readwrite.json_graph

from sklearn.base import BaseEstimator, TransformerMixin

from dyneusr import datasets
from dyneusr import visuals
from dyneusr import tools



class DyNeuGraph(BaseEstimator, TransformerMixin):

    def __init__(self, **params):
        """ DyNeuGraph
        
        Parameters
        ----------

        Usage
        -----

            # Fit DyNeuGraph
            dyneuG = dy.DyNeuGraph()
            dyneuG.fit(X)

            # Transform Mapper graph
            tcm = dyneuG.transform(graph)       # [T x N] Time-Node adjacency
    
            # Annotate
            dyneuG.annotate(microstates=dict(color='blue'))

            # Visualize
            dyneuG.visualize()
        """
        self.cache_ = dict(params)
        self.fit(**params)


    def cache(self, *args, **kwargs):
        """ Cache misc. data here

        TODO: hacky, replace with getter/setter
        """
        # return value, if arg is passed
        if len(args):
            for _ in args:
                if self.cache_.get(_) is None:
                    continue
                return self.cache_.get(_)
            return None
        # otherwise, update
        self.cache_.update(kwargs)
        return self.cache_



    def fit(self, G=None, X=None, y=None, node_data=dict(), edge_data=dict(), G_data=False, **kwargs):
        """ Fit to G.

        Usage
        -----
            # Transform Mapper graph
            dyneuG.fit(graph, y=meta)       

        """
        # save inputs
        self.G_input_ = G
        self.X_ = X 
        self.y_ = y

        # check graph
        if isinstance(G, nx.Graph):
            g = G.copy()
            G = nx.readwrite.json_graph.node_link_data(g)
        elif G is None:
            G = {}

        # copy graph, override defaults
        G = dict(dict(nodes={}, links={}), **G)

        # states, microstates
        node_ids = np.unique([n for i,n in enumerate(G['nodes'])])
        data_ids = np.unique([_ for n,d in G['nodes'].items() for _ in d])
        if y is not None:
            if isinstance(y, pd.DataFrame):
                data_ids = np.sort(y.reset_index().index)
            else:
                data_ids = np.arange(len(y))

        # process graph
        G = tools.process_graph(G, meta=y, **kwargs)
        A, M, TCM = tools.extract_matrices(G, index=data_ids)

        # create graph from TCM
        if G_data is True:
            G_data = nx.MultiGraph(TCM)
            nx.relabel_nodes(G_data, dict(zip(G_data, data_ids)))
            nx.set_node_attributes(G_data, dict(zip(data_ids, y)), 'group')

        # annotate any additional node datahere
        if node_data is not None:
            for k in node_data:
                nx.set_node_attributes(G, dict(node_data[k]), k)
        if edge_data is not None:
            for k in edge_data:
                   nx.set_edge_attributes(G, dict(edge_data[k]), k)

        # mixture of connected TRs, for each TR 
        mixtures = [_.nonzero()[0] for _ in TCM]


        # store variables
        self.G_ = G
        self.G_data_ = G_data
        self.node_ids_ = node_ids 
        self.data_ids_ = data_ids
        self.adj_ = A               # node adjacency matrix
        self.map_ = M             # node attr matrix
        self.tcm_ =  TCM                # temporal connectivity
        self.mixtures_ = mixtures

        # some new conventions
        #self.G = self.G_
        #self.A = self.adj_
        #self.M = self.map_
        #self.TCM = self.tcm_
        return self


    @property
    def G(self):
        return self.G_

    @property
    def A(self):
        return self.adj_
    
    @property
    def M(self):
        return self.map_
    
    @property
    def TCM(self):
        return self.tcm_
    

    def inverse_transform(self, G, y=None):
        """ Inverse transform of G 
        """
        return self.G_inverse_


    def fit_transform(self, G, y=None):
        """ Transform Mapper graph into populations.

        Usage
        -----
            # Transform Mapper graph
            assigns = dyneuG.fit_transform(graph, y=meta)      

        """
        # fit
        self.fit(G=G, y=y)
        return self.tcm_


    def sample(self, X, y=None):
        """ Predict Mixture Models using data based on TCM
        """
        mixtures_img = None
        # TODO:
        self.mixtures_img_ = mixtures_img
        return self.mixtures_img_    


    def transform(self, X, y=None):
        """ Transform X into TCM.

        Usage
        -----
            # Transform Mapper graph
            TCM = dyneuG.transform(X, y=meta)      

        """
        A, M, TCM = tools.extract_matrices(self.G_)

        # create graph from TCM
        if G_data is True:
            G_data = nx.MultiGraph(TCM)
            nx.relabel_nodes(G_data, dict(zip(G_data, data_ids)))
            nx.set_node_attributes(G_data, dict(zip(data_ids, y)), 'group')


        # mixture of connected TRs, for each TR 
        mixtures = [_.nonzero()[0] for _ in TCM]

        # store variables
        self.G_data_ = G_data
        self.adj_ = A               # node adjacency matrix
        self.map_ = M             # node attr matrix
        self.tcm_ =  TCM                # temporal connectivity
        self.mixtures_ = mixtures
        return self.tcm_


    def annotate(self, **kwargs):
        """ Annotate graph with data.

        TODO: not sure how to do this...
        """
        G_, annotations = visuals.annotate(self.G_, **kwargs)
        
        # save
        self.G_ = G_
        self.annotations_ = annotations
        return self


    def annotate_nodes(self, **kwargs):
        """    Set node attributes from dictionary of nodes and values. 

        Parameters
        ----------
            name: string
                Attribute name
            values: dict
                Dictionary of attribute values keyed by node.
            kwargs: dict
                Dictionary of attribute values keyed by name.
        
        Examples
        --------
            dG.annotate_nodes(color='blue')


        """
        for name, values in kwargs.items():
            if isinstance(values, np.ndarray):
                values = list(values)
            elif isinstance(values, dict):
                values = values.keys()
            elif not isinstance(values, list):
                values = [values for _ in self.G_]
            values = {n:value for n,value in zip(self.G_,values)}
            nx.set_node_attributes(self.G_, values, name)
        #nx.set_node_attributes(self.G_, name, values)
        
        # save
        return self


    def annotate_members(self, **kwargs):
        """    Set node attributes from dictionary of members and values. 

        Parameters
        ----------
            name: string
                Attribute name
            values: dict
                Dictionary of attribute values keyed by node.
            kwargs: dict
                Dictionary of attribute values keyed by name.

        Examples
        --------
            dG.annotate_node(image=data_imgs)

        """
        # TODO: map_from_nodes
        for name, values in kwargs.items():
            if isinstance(values, np.ndarray):
                values = list(values)
            elif isinstance(values, dict):
                values = values.keys()
            elif not isinstance(values, list):
                values = [values]
            # map attrs to data_ids
            values = {n:value for n,value in zip(self.G_data_,values)}
            nx.set_node_attributes(self.G_data_, values, name)

        # save
        return self


    def annotate_graph(self, **kwargs):
        """ Annotate graph with data.
        """
        self.G_.graph.update(**kwargs)
        return self


    def visualize(self, path_html='index.html', json_graph=None, color_functions=None, custom_data=None, plot_tcm=False, static=True, show=False, port=None, **kwargs):
        """ Visualize DyNeuGraph.

        TODO: this needs some work...
        """
        # color functions
        if color_functions is not None:
            if not isinstance(color_functions, dict):
                color_functions = dict(default=color_functions)
            self.G.graph.update(color = color_functions)

        

        # format html
        if isinstance(custom_data, dict):
            self.annotate_graph(**custom_data)
        
        # to node_link
        self.node_link_data_ = dict(nx.readwrite.json_graph.node_link_data(self.G))

        # [1] plot TCM
        if plot_tcm:
            figs = visuals.plot_temporal_matrix(self.tcm_, y=None, show=show, **kwargs)

        # [2] visualize force
        HTTP = visuals.visualize_force(self.node_link_data_, path_html=path_html, static=static, show=show, port=port, **kwargs)
        self.HTTP = HTTP

        # [3] open in browser (optional)
        #if show is True:
        #    self.show()
        return self


    def show(self):
        """ Open the HTTP url using webbrowser

        """
        try:
            import webbrowser
            webbrowser.open(self.HTTP.url)
        except ImportError as e:
            print(e)
            print('Hint: requires Python webbrowser module...')
        return self


    def add_custom_layout(self, pos, name='custom'):
        """Add a user-defined list of positions

        Parameters
        ----------
        pos : dict, array-like, callable
            Dictionary, list, or function for defining the
            coordinate positions of nodes in the graph.

        name : str
            Name of the custom_layout.

        Examples
        --------
        >>> dG.add_custom_layout(pca, name='pca')
        >>> dG.add_custom_layout(nx.spectral_layout, 'spectral')

        """
        def check_node_layout(pos):
            # check for layout functions
            if callable(pos):
                pos = dict(pos(self.G))
            
            # check for positions defined for each member
            if len(pos) != len(self.G):
                pos = {n: np.mean(pos[_], axis=0) for n,_ in self.G.nodes('members')}
            
            # make sure we have a dictionary with nodes as the keys
            if not isinstance(pos, dict):
                pos = {n: pos[i] for i,n in enumerate(self.G)}
            
            # normalize
            from sklearn.preprocessing import MinMaxScaler
            arr = np.stack([np.array(pos[n]) for n in pos])
            arr = MinMaxScaler().fit_transform(arr)
            arr = arr.round(6).astype(float)

            # split components
            arrs = {'{} {}'.format(name, i+1): 
                    _ for i,_ in enumerate(arr.T)}

            # convert all arrays to lists, return
            pos = {k: {n:arrs[k][i] for i,n in enumerate(pos)} for k in arrs}
            return pos

        # make sure we have positions defined for each node
        pos = check_node_layout(pos)

        # get current custom_layouts, then update
        custom_layouts = self.G.graph.get('custom_layouts', {})
        custom_layouts.update(**pos)

        # reset custom layouts
        self.G.graph.update(custom_layouts=custom_layouts)
        return self

